import numpy as np
import triangle as tr
from scipy.spatial import Delaunay as delaunay
from shapely.geometry import Polygon as ShapelyPolygon
from shapely.geometry import LineString as ShapelyLine
from shapely.geometry import Point as ShapelyPoint

class Mesh:
    def __init__(self, x, y, conmat):
        self.x = x
        self.y = y
        self.conmat = conmat

    def getNeiElements(self):
        nn = len(self.x)
        ne = len(self.conmat[:,0])
        neiElements = []
        for i in range(nn):
            ele = []
            for j in range(ne):
                if self.conmat[j,0] == i or self.conmat[j,1] == i or self.conmat[j,2] == i:
                    ele.append(j)
            neiElements.append(ele)
        self.neiElements = neiElements

    def getNeiNodes(self):
        nn = len(self.x)
        neiNodes = []
        for i in range(nn):
            nod = []
            for j in self.neiElements[i]:
                for k in self.conmat[j]:
                    if k != i: 
                        nod.append(k)
            nod = np.unique(np.array(nod))
            nod = nod.tolist()
            neiNodes.append(nod)
        self.neiNodes = neiNodes
        
def prepareGeometryForMeshing(shapes):
    # Finding all child shapes
    for i in range(len(shapes)): # for all shapes,
        if shapes[i].isClosed: # if the current shape is closed, it may contain a child.
            for j in range(len(shapes)): # for all potential child shapes,
                if i != j: # skip current shape as potential child.
                    polygon = ShapelyPolygon(shapes[i].boundaryNodes) # consider current shape as polygon,
                    if shapes[j].shape == "point":
                        point = ShapelyPoint(shapes[j].boundaryNodes) # and take one point form potential child shape.
                    else:
                        point = ShapelyPoint(shapes[j].boundaryNodes[0]) # and take one point form potential child shape.
                    if polygon.contains(point): # if one point of potential child shape is inside current shape,
                        shapes[i].child.append(j) # potential child shape is a child of current shape. Note: all the shapes must be non-contacting.

    # Finding all parent shapes
    for i in range(len(shapes)): # for all shapes,
        for j in range(len(shapes)): # for all potential parent shapes,
            if i != j: # skip current shape as potential parent.
                if shapes[j].isClosed: # if the potential shape is closed, it may be a parent.
                    polygon = ShapelyPolygon(shapes[j].boundaryNodes) # consider potential parent shape as polygon,
                    if shapes[i].shape == "point":
                        point = ShapelyPoint(shapes[i].boundaryNodes) # and take one point from current shape.
                    else:
                        point = ShapelyPoint(shapes[i].boundaryNodes[0]) # and take one point from current shape.
                    if polygon.contains(point): # if one point of current shape is inside the potential parent shape,
                        shapes[i].parent.append(j) # potential parent shape is the parent of current shape. Note: all the shapes must be non-contacting.

    for i in range(len(shapes)):
        shapes[i].parent = np.array(shapes[i].parent)
        shapes[i].child = np.array(shapes[i].child)

    # Finding immediate child shapes
    listOfShapesToRemove = []
    for i in range(len(shapes)):
        shapesToRemove = []
        for j in range(len(shapes[i].child)):
            for k in range(len(shapes[shapes[i].child[j]].child)):
                shapesToRemove.append(shapes[shapes[i].child[j]].child[k])
        listOfShapesToRemove.append(shapesToRemove)

    for i in range(len(shapes)):
        for j in range(len(listOfShapesToRemove[i])):
            shapes[i].child = np.delete(shapes[i].child, np.where(shapes[i].child == listOfShapesToRemove[i][j]))#, axis=0)

    # Finding immediate parent shapes
    listOfShapesToRemove = []
    for i in range(len(shapes)):
        shapesToRemove = []
        for j in range(len(shapes[i].parent)):
            for k in range(len(shapes[shapes[i].parent[j]].parent)):
                shapesToRemove.append(shapes[shapes[i].parent[j]].parent[k])
        listOfShapesToRemove.append(shapesToRemove)
        
    for i in range(len(shapes)):
        for j in range(len(listOfShapesToRemove[i])):
            shapes[i].parent = np.delete(shapes[i].parent, np.where(shapes[i].parent ==listOfShapesToRemove[i][j]))#, axis=0)

    # Labelling if a shape is child or parent
    for i in range(len(shapes)):
        if len(shapes[i].child) == 0:
            shapes[i].isParent = False
        else:
            shapes[i].isParent = True

    for i in range(len(shapes)):
        if len(shapes[i].parent) == 0:
            shapes[i].isChild = False
        else:
            shapes[i].isChild = True

def generateNodes(shapes, meshingMethod, maxAngle, maxArea, crse):
    # Generate interior free nodes
    if meshingMethod == "ruppert":
        for i in range(len(shapes)):
            if shapes[i].isClosed:
                meshShapeRuppert(i, shapes, crse)
            else:
                shapes[i].freeNodes = np.array([])
    elif meshingMethod == "triangle":
        meshShapesTriangle(shapes,maxAngle, maxArea)

def meshShapesTriangle(shapes, maxAngle, maxArea):
    iv, fv = triangleMesh(shapes, maxAngle, maxArea)
    initNodes = iv.shape[0]
    finalNodes = fv.shape[0]
    for shape in shapes:
        shape.freeNodes = np.empty((0,2))
    for i in range(initNodes,finalNodes):
        extraVertex = ShapelyPoint(fv[i][0],fv[i][1])
        for shape in shapes:
            if shape.isClosed:
                shapePolygon = ShapelyPolygon(shape.boundaryNodes.tolist())
                # if extraVertex is on boundary
                if not shapePolygon.contains(extraVertex) and shapePolygon.covers(extraVertex):
                    shape.boundaryNodes = np.vstack([shape.boundaryNodes, np.array([fv[i][0],fv[i][1]])])
                # if extraVertex is inside
                if shapePolygon.contains(extraVertex):
                    vertexOutside = True
                    for j in shape.child:
                        childShape = shapes[j]
                        childPolygon = ShapelyPolygon(childShape.boundaryNodes.tolist())
                        if childPolygon.covers(extraVertex):
                            vertexOutside = False
                            break
                    if vertexOutside:
                        shape.freeNodes = np.vstack([shape.freeNodes, np.array([fv[i][0],fv[i][1]])])
            elif shape.shape != 'point':
                lineString = ShapelyLine(shape.boundaryNodes.tolist())
                if lineString.contains(extraVertex):
                    shape.boundaryNodes = np.vstack([shape.boundaryNodes, np.array([fv[i][0],fv[i][1]])])

def triangleMesh(shapes, maxAngle = "", maxArea = ""):
    nov = 0
    ver = np.empty((0,2))
    seg = np.empty((0,2), dtype=int)
    for shape in shapes:
        ver = np.vstack([ver, shape.boundaryNodes])
        if shape.shape != 'point':
            seg = np.vstack([seg, np.asarray(shape.segments, dtype=int)+ nov])
        nov = ver.shape[0]
    pslg = dict(vertices=ver, segments=seg)
    stri = "pq"+maxAngle+"a"+maxArea+"A"
    tri = tr.triangulate(pslg, stri)
    return ver, tri['vertices']

def meshShapeRuppert(ind, shapes, crse): # Consider skipping sengment initialisation in case of Point
    # Parent points
    parentPoints = shapes[ind].boundaryNodes
    # Child points
    childPoints = []
    for i in range(len(shapes[ind].child)):
        childPoints.append(shapes[shapes[ind].child[i]].boundaryNodes.tolist())
    
    # Parent polygon
    parentPolygon = ShapelyPolygon(parentPoints)
    # Child polygons
    childPolygons = []
    for i in range(len(shapes[ind].child)):
        if shapes[shapes[ind].child[i]].isClosed:
            childPolygons.append(ShapelyPolygon(childPoints[i]))

    # Parent segments
    segments = []
    for i in range(len(parentPoints)-1):
        segments.append([i, i+1])
    segments.append([len(parentPoints)-1, 0])
    # Child segments
    segInd = len(parentPoints)
    for i in range(len(shapes[ind].child)):
        for j in range(len(childPoints[i])-1):
            if shapes[shapes[ind].child[i]].shape != "point":
                segments.append([segInd+j, segInd+j+1])
        if shapes[shapes[ind].child[i]].isClosed:
            segments.append([segInd+j+1, segInd])
        if shapes[shapes[ind].child[i]].shape == 'point':
            segInd = segInd + 1
        else:
            segInd = segInd + len(childPoints[i])

    # Points
    points = parentPoints

    for i in range(len(childPoints)):
        points = np.vstack((points,childPoints[i]))

    dt = delaunay(points)
    cm = dt.simplices  # Connectivity matrix
    x = points[:, 0]
    y = points[:, 1]
    ix = x.copy()
    iy = y.copy()

    elementsToRemove = []
    for i in range(len(cm)):
        cot = ShapelyPoint((x[cm[i][0]]+x[cm[i][1]]+x[cm[i][2]])/3, (y[cm[i][0]]+y[cm[i][1]]+y[cm[i][2]])/3)
        if not parentPolygon.contains(cot):
            elementsToRemove.append(i)
        for childPolygon in childPolygons:
            if childPolygon.contains(cot):
                elementsToRemove.append(i)
    cm = np.delete(cm, elementsToRemove, 0)
    st = getSharpTriangles(x, y, cm, crse)

    while len(st) != 0:
        for element in st:
            ex = x[cm[element]]
            ey = y[cm[element]]
            cc = circumCenter(ex, ey)

            isNotEcroached = segmentsEncroached(points, segments, cc)
            cirCen = ShapelyPoint(cc[0], cc[1])
            if parentPolygon.contains(cirCen) and isNotEcroached:
                pointOutsideChild = True
                for i in range(len(childPolygons)):
                    if childPolygons[i].contains(cirCen):
                        pointOutsideChild = False
                        break
                if pointOutsideChild:
                    points = np.vstack([points, cc])
                    dt = delaunay(points)
                    cm = dt.simplices # Connectivity matrix
                    x = points[:,0]
                    y = points[:,1]

                    elementsToRemove = []
                    for i in range(len(cm)):
                        cot = ShapelyPoint((x[cm[i][0]]+x[cm[i][1]]+x[cm[i][2]])/3, (y[cm[i][0]]+y[cm[i][1]]+y[cm[i][2]])/3)
                        if not parentPolygon.contains(cot):
                            elementsToRemove.append(i)
                        for childPolygon in childPolygons:
                            if childPolygon.contains(cot):
                                elementsToRemove.append(i)
                    cm = np.delete(cm, elementsToRemove, 0)
                    break
        stNew = getSharpTriangles(x, y, cm, crse)
        if stNew == st:
            break
        else:
            st = stNew

    fx = x[len(ix):len(x)-1]
    fy = y[len(iy):len(y)-1]
    freeNodes = []
    for i in range(len(fx)):
        freeNodes.append([fx[i],fy[i]])
    freeNodes = np.array(freeNodes)
    shapes[ind].freeNodes = freeNodes

def circumCenter(ex, ey):
    Ax, Ay, Bx, By, Cx, Cy = ex[0], ey[0], ex[1], ey[1], ex[2], ey[2]
    D = 2*(Ax*(By-Cy)+Bx*(Cy-Ay)+Cx*(Ay-By))
    cx = ((Ax*Ax+Ay*Ay)*(By-Cy)+(Bx*Bx+By*By)*(Cy-Ay)+(Cx*Cx+Cy*Cy)*(Ay-By))/D
    cy = ((Ax*Ax+Ay*Ay)*(Cx-Bx)+(Bx*Bx+By*By)*(Ax-Cx)+(Cx*Cx+Cy*Cy)*(Bx-Ax))/D
    return cx, cy

def circumRadiusToMinSideRatio(ex, ey):
    a, b, c = np.sqrt((ex[1]-ex[0])**2+(ey[1]-ey[0])**2), np.sqrt((ex[2]-ex[1])**2+(ey[2]-ey[1])**2), np.sqrt((ex[2]-ex[0])**2+(ey[2]-ey[0])**2)
    minSide = min([a, b, c])
    cr = (a*b*c)/(np.sqrt((a + b + c)*(b + c - a)*(c + a - b)*(a + b - c))) # Circumradius
    return cr/minSide

def getSharpTriangles(x, y, cm, B):
    st = []
    for i in range(len(cm)):
        element = cm[i]
        ex = x[element]
        ey = y[element]
        circumRatio = circumRadiusToMinSideRatio(ex, ey)
        if circumRatio > B:
            st.append(i)
    return st

def encroached(p1, p2, p):
    m = (np.array(p1)+np.array(p2))/2
    r = np.sqrt((p1[0]-m[0])**2+(p1[1]-m[1])**2)
    d = np.sqrt((p[0]-m[0])**2+(p[1]-m[1])**2)
    if d > r:
        return False
    else:
        return True

def segmentsEncroached(points, segments, cc):
    isNotEncroached = True
    for i in range(len(segments)):
        segmentCentre = [(points[segments[i][0]][0]+points[segments[i][1]][0])/2 ,(points[segments[i][0]][1]+points[segments[i][1]][1])/2]
        segmentRadius = np.sqrt((points[segments[i][0]][0]-points[segments[i][1]][0])**2+(points[segments[i][0]][1]-points[segments[i][1]][1])**2)/2
        distanceFromCentre = np.sqrt((cc[0]-segmentCentre[0])**2+(cc[1]-segmentCentre[1])**2)
        if distanceFromCentre < segmentRadius:
            isNotEncroached = False
            break
    return isNotEncroached

def createMesh(shapes):

    x = np.array([])
    y = np.array([])
    nn = -1

    for shape in shapes:
        if shape.shape == "point":
            shape.boundaryList = np.array([nn+1])
            nn = nn+1
            shape.freeList = np.array([])
            x =  np.concatenate((x, [shape.boundaryNodes[0]]))
            y =  np.concatenate((y, [shape.boundaryNodes[1]]))

        elif shape.shape == "line" or shape.shape == "polyline":
            shape.boundaryList = np.arange(nn+1, nn+len(shape.boundaryNodes[:, 0])+1, 1)
            nn = shape.boundaryList[-1]
            shape.freeList = np.array([])
            x =  np.concatenate((x, shape.boundaryNodes[:, 0]))
            y =  np.concatenate((y, shape.boundaryNodes[:, 1]))

        else:

            shape.boundaryList = np.arange(nn+1, nn+len(shape.boundaryNodes)+1, 1)
            nn = shape.boundaryList[-1]
            if len(shape.freeNodes) == 0:
                shape.freeList = np.array([])
            else:
                shape.freeList = np.arange(nn+1, nn+len(shape.freeNodes)+1, 1)
                nn = shape.freeList[-1]
            x =  np.concatenate((x, shape.boundaryNodes[:, 0], shape.freeNodes[:, 0]))
            y =  np.concatenate((y, shape.boundaryNodes[:, 1], shape.freeNodes[:, 1]))

    xy = np.transpose([x, y])
    dt = delaunay(xy)
    mesh = Mesh(x, y, dt.simplices)
    
    # Generate element list in each shape
    ne = len(dt.simplices[:,0])
    parentPolygons = []
    childPolygons = []
    for shape in shapes:
        shape.elementList = []
        if shape.isClosed:
            parentPolygons.append(ShapelyPolygon(shape.boundaryNodes))
            childPolygon = []
            for i in range(len(shape.child)):
                if shapes[shape.child[i]].shape == "point":
                    childPolygon.append([])
                else:
                    childPolygon.append(ShapelyPolygon(shapes[shape.child[i]].boundaryNodes))
            childPolygons.append(childPolygon)
        else:
            parentPolygons.append([])
            childPolygons.append([])

    cm = dt.simplices
    for i in range(ne):
        ex = x[cm[i]]
        ey = y[cm[i]]
        cc = circumCenter(ex, ey)
        point = ShapelyPoint(cc[0], cc[1])

        for j in range(len(shapes)):
            if shapes[j].isClosed:
                parentPolygon = parentPolygons[j]
                childPolygon = childPolygons[j]
                if parentPolygon.contains(point):
                    pointOutsideChild = True
                    for k in range(len(childPolygon)):
                        if shapes[shapes[j].child[k]].shape != 'point':
                            if childPolygon[k].contains(point):
                                pointOutsideChild = False
                                break

                    if pointOutsideChild:
                        shapes[j].elementList.append(i)

    # Elements
    mesh.getNeiElements()
    # Nodes
    mesh.getNeiNodes()

    return mesh

def assignSources(shapes, mesh):
    x = mesh.x
    y = mesh.y
    nn = len(x)
    # nodeSource is for both plotting and matrix assignment
    # elementSource may serve some purpose in different physics
    nodeSource = np.zeros((nn,1))
    ne = len(mesh.conmat[:,0])
    elementSource = np.zeros((ne,1))

    for shape in shapes:
        if not shape.isClosed:
            nodes = shape.boundaryList
            for node in nodes:
                nodeSource[node] = shape.source/len(nodes)
                elements = mesh.neiElements[node]
                for element in elements:
                    elementSource[element] = elementSource[element] + shape.source/len(elements)
        else:
            elements = shape.elementList
            for element in elements:
                ex = []
                ey = []
                for node in mesh.conmat[element]:
                    ex.append(x[node])
                    ey.append(y[node])
                p = [ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]]
                q = [ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]]
                ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
                elementSource[element] = shape.source*ea
            for element in elements:
                for node in mesh.conmat[element]:
                    nodeSource[node] = nodeSource[node] + elementSource[element]/3
    mesh.elementSource = elementSource            
    mesh.nodeSource = nodeSource
    
def assignMaterials(shapes, mesh):
    # nodeMaterial may be used for plotting
    # elementMaterial is for matrix computation
    x = mesh.x
    nn = len(x)
    nodeMaterial = np.ones((nn,1))
    ne = len(mesh.conmat[:,0])
    elementMaterial = np.ones((ne,1))

    for shape in shapes:
        if not shape.isClosed:
            nodes = shape.boundaryList
            for node in nodes:
                nodeMaterial[node] = shape.material
        else:
            nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
            for node in nodes:
                nodeMaterial[node] = shape.material
        elements = shape.elementList
        for element in elements:
            elementMaterial[element] = shape.material
    mesh.nodeMaterial = nodeMaterial
    mesh.elementMaterial = elementMaterial




import numpy as np

class Geometry:

    def __init__(self, **kwargs):
        self.stroke = kwargs.get('stroke', "black")
        self.strokeWidth = kwargs.get('strokeWidth', "1")
        self.fill = kwargs.get('fill', "transparent")
        self.non = kwargs.get('non', "0")
        self.relativeDensity = kwargs.get('relativeDensity', "[1, 1]")
        self.scale = kwargs.get('scale', "1")
        self.rotate = kwargs.get('rotate', "0")
        self.pivot = kwargs.get('pivot', "itself")
        self.pivotAngle = kwargs.get('pivotAngle', "0")
        self.parent = []
        self.child = []
        self.isParent = True
        self.isChild = True
        self.isHole = kwargs.get('isHole', False) # Work on it later
        self.source = 0
        self.material = 1




class Circle(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "circle"
        self.cx = kwargs['cx']
        self.cy = kwargs['cy']
        self.r = kwargs['r']
        self.non = kwargs['non']
        self.isClosed = True

        r = self.r
        th = np.linspace(0,2*np.pi,self.non +1)
        x = r*np.cos(th[0:-1])
        y = r*np.sin(th[0:-1])

        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.cx], [self.cy]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.cx, self.cy])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        segments.append([len(boundaryNodes)-1, 0])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))




class Ellipse(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "ellipse"
        self.cx = kwargs['cx']
        self.cy = kwargs['cy']
        self.rx = kwargs['rx']
        self.ry = kwargs['ry']
        self.non = kwargs['non']
        self.isClosed = True

        a = self.rx
        b = self.ry
        th = np.linspace(0,2*np.pi,self.non + 1)
        x = a*np.cos(th[0:-1])
        y = b*np.sin(th[0:-1])

        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.cx], [self.cy]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.cx, self.cy])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        segments.append([len(boundaryNodes)-1, 0])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))




class Line(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "line"
        self.x1 = kwargs['x1']
        self.x2 = kwargs['x2']
        self.y1 = kwargs['y1']
        self.y2 = kwargs['y2']
        self.non = kwargs['non']
        self.isClosed = False

        non = self.non
        if self.x1 == self.x2:
            x = np.arange(int(non))*0 + self.x1
            y = np.arange(self.y1, self.y2, (self.y2-self.y1)/(non-1))
            y = np.hstack([y, self.y2])
        elif self.y1 == self.y2:
            x = np.arange(self.x1, self.x2, (self.x2-self.x1)/(non-1))
            x = np.hstack([x, self.x2])
            y = np.arange(int(non))*0 + self.y1
        else:
            x = np.arange(self.x1, self.x2, (self.x2-self.x1)/(non-1))
            x = np.hstack([x, self.x2])
            y = np.arange(self.y1, self.y2, (self.y2-self.y1)/(non-1))
            y = np.hstack([y, self.y2])

        x = x-x[0]
        y = y-y[0]
        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.x1], [self.y1]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.x1, self.y1])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        
    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))




class Point(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "point"
        self.point = kwargs['point']
        self.isClosed = False

        # Scale the points
        v0 = (np.array(self.point)-np.array(self.point))*float(self.scale)
        v0 = np.transpose(v0)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.point[0]], [self.point[1]]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array(self.point)
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        plt.plot(v[0], v[1], "*", color = self.stroke, markersize = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        plt.plot(v[0], v[1], ".", color = self.stroke, markersize = float(self.strokeWidth))




class Polygon(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "polygon"
        self.points = kwargs['points']
        self.isClosed = True

        # Scale the points
        v0 = (np.array(self.points)-np.array(self.points[0]))*float(self.scale)
        v0 = np.transpose(v0)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.points[0][0]], [self.points[0][1]]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array(self.points[0])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        segments.append([len(boundaryNodes)-1, 0])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))




class Polyline(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "polyline"
        self.points = kwargs['points']
        self.isClosed = False

        # Scale the points
        v0 = (np.array(self.points)-np.array(self.points[0]))*float(self.scale)
        v0 = np.transpose(v0)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.points[0][0]], [self.points[0][1]]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array(self.points[0])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        
        


class Rectangle(Geometry):

    def __init__(self, **kwargs):
        Geometry.__init__(self, **kwargs)
        self.shape = "rect"
        self.x = kwargs['x']
        self.y = kwargs['y']
        self.rx = kwargs.get('rx', "0")
        self.ry = kwargs.get('ry', "0")
        self.width = kwargs['width']
        self.height = kwargs['height']
        self.non = kwargs['non']
        self.isClosed = True

        w = self.width
        h = self.height
        non = self.non
        edge1x = np.arange(-w/2, w/2, w/non)
        edge1y = np.arange(len(edge1x))*0 - h/2
        edge2y = np.arange(-h/2, h/2, h/(non*h/w))
        edge2x = np.arange(len(edge2y))*0 + w/2
        
        edge3x = np.arange(w/2, -w/2, -w/non)
        edge3y = np.arange(len(edge3x))*0 + h/2
        edge4y = np.arange(h/2, -h/2, -h/(non*h/w))
        edge4x = np.arange(len(edge4y))*0 - w/2
        
        x = np.hstack([edge1x, edge2x, edge3x, edge4x])
        y = np.hstack([edge1y, edge2y, edge3y, edge4y])

        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.x], [self.y]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.x, self.y])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        segments.append([len(boundaryNodes)-1, 0])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))








# Import packages
import joblib
import numpy as np

def vectorField(par, location, isNode = False, nodeNumber = -1):

    if par == []:
        shapes = joblib.load("Data/geometry.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        shapes = joblib.load(stri)
        stri = f'Data/mesh_{par}.sav'
        mesh = joblib.load(stri)
        stri = f'Data/solver_{par}.sav'
        solver = joblib.load(stri)

    x = mesh.x
    y = mesh.y
    tri = mesh.conmat
    u = solver.u.copy()
    if isinstance(location, int):
        if shapes[location].shape == 'point':
            nodeNumber = shapes[location].boundaryList[0]
            return(vectorField(par, np.array(shapes[location].position), True, nodeNumber))
        elif shapes[location-1].shape == 'line' and shapes[location].shape == 'line' and shapes[location+1].shape == 'line':
            l1 = shapes[location-1].boundaryList
            l2 = shapes[location].boundaryList
            l3 = shapes[location+1].boundaryList
            field = []
            for i in range(1,len(l2)-1):
                Exp = (u[l2[i]]-u[l2[i+1]])/(np.sqrt((x[l2[i]] - x[l2[i+1]])**2+(y[l2[i]] - y[l2[i+1]])**2))
                Exn = (u[l2[i-1]]-u[l2[i]])/(np.sqrt((x[l2[i-1]] - x[l2[i]])**2+(y[l2[i-1]] - y[l2[i]])**2))
                Ex = (Exp+Exn)/2
                Eyp = (u[l2[i]]-u[l1[i]])/(np.sqrt((x[l2[i]] - x[l1[i]])**2+(y[l2[i]] - y[l1[i]])**2))
                Eyn = (u[l3[i]]-u[l2[i]])/(np.sqrt((x[l3[i]] - x[l2[i]])**2+(y[l3[i]] - y[l2[i]])**2))
                Ey = (Eyp+Eyn)/2
                field.append([Ex, Ey])

            if solver.name == "magnetostatic":
                thr = 90
                R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
                [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
                for i in range(len(field)):
                    field[i] = np.dot(R,field[i])

            return field
        else:
            raise Exception("Field along the line cannot be obtained for the selected shape", location)

    elif isinstance(location, np.ndarray):

        if isNode:
            field = [0, 0]
            neiElements = mesh.neiElements[nodeNumber]
            elementfields = []
            for element in neiElements:
                # A = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
                # for k in range(3):
                #     A[k] = [1, x[tri[element][k]], y[tri[element][k]]]
                # b = [[u[tri[element][0]][0]], [u[tri[element][1]][0]], [u[tri[element][2]][0]]]
                # X = np.linalg.lstsq(A, b)[0]
                # elementfield = [-X[1][0], -X[2][0]]
                A = abs(area(x[tri[element][0]], y[tri[element][0]], x[tri[element][1]], y[tri[element][1]], x[tri[element][2]], y[tri[element][2]]))
                Ex = (-1/(2*A))*((y[tri[element][1]]-y[tri[element][2]])*u[tri[element][0]]+(y[tri[element][2]]-y[tri[element][0]])*u[tri[element][1]]+(y[tri[element][0]]-y[tri[element][1]])*u[tri[element][2]])                
                Ey = (-1/(2*A))*((x[tri[element][2]]-x[tri[element][1]])*u[tri[element][0]]+(x[tri[element][0]]-x[tri[element][2]])*u[tri[element][1]]+(x[tri[element][1]]-x[tri[element][0]])*u[tri[element][2]])                
                elementfield = [Ex, Ey]
                elementfields.append(elementfield)
            elementfields = np.array(elementfields)
            field = sum(elementfields)/len(elementfields)

        else:
            field = []
            noOfRows = location.shape[0]
            for i in range(noOfRows):
                field.append([0, 0])

            for j in range(noOfRows):
                isInsideGlobalBoundary = False
                for i in range(len(tri)):
                    if isInside(x[tri[i][0]], y[tri[i][0]], x[tri[i][1]], y[tri[i][1]], x[tri[i][2]], y[tri[i][2]], location[j][0], location[j][1]):
                        theElement = i
                        isInsideGlobalBoundary = True
                        break

                if isInsideGlobalBoundary:
                    # A = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
                    # for k in range(3):
                    #     A[k] = [1, x[tri[theElement][k]], y[tri[theElement][k]]]
                    # b = [[u[tri[theElement][0]][0]], [u[tri[theElement][1]][0]], [u[tri[theElement][2]][0]]]
                    # X = np.linalg.lstsq(A, b)[0]
                    # field[j] = [-X[1][0], -X[2][0]]
                    A = abs(area(x[tri[theElement][0]], y[tri[theElement][0]], x[tri[theElement][1]], y[tri[theElement][1]], x[tri[theElement][2]], y[tri[theElement][2]]))
                    Ex = (-1/(2*A))*((y[tri[theElement][1]]-y[tri[theElement][2]])*u[tri[theElement][0]]+(y[tri[theElement][2]]-y[tri[theElement][0]])*u[tri[theElement][1]]+(y[tri[theElement][0]]-y[tri[theElement][1]])*u[tri[theElement][2]])                
                    Ey = (-1/(2*A))*((x[tri[theElement][2]]-x[tri[theElement][1]])*u[tri[theElement][0]]+(x[tri[theElement][0]]-x[tri[theElement][2]])*u[tri[theElement][1]]+(x[tri[theElement][1]]-x[tri[theElement][0]])*u[tri[theElement][2]])                
                    field[j] = [Ex, Ey]

                else:
                    field[j] = [[0, 0]]
                    raise Exception("Position", location[j], "is out of Global Boundary")

        if solver.name == "magnetostatic":
            thr = 90
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            field = np.dot(R,field)

        return field

def area(x1, y1, x2, y2, x3, y3):
    return abs((x1 * (y2 - y3) + x2 * (y3 - y1)+ x3 * (y1 - y2))/ 2.0)

def isInside(x1, y1, x2, y2, x3, y3, x, y):
    A = area(x1, y1, x2, y2, x3, y3)
    A1 = area(x, y, x2, y2, x3, y3)
    A2 = area(x1, y1, x, y, x3, y3)
    A3 = area(x1, y1, x2, y2, x, y)
    AT = A1 + A2 + A3
    if(AT < A*1.00001): # To avoid floating point error
        return True
    else:
        return False

def force(par, shapeNumber):

    if par == []:
        shapes = joblib.load("Data/geometry.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        shapes = joblib.load(stri)
        stri = f'Data/mesh_{par}.sav'
        mesh = joblib.load(stri)
        stri = f'Data/solver_{par}.sav'
        solver = joblib.load(stri)

    x = mesh.x
    y = mesh.y
    nodeSource = mesh.nodeSource
    inducedNodeSource = solver.bi

    nodes = np.concatenate((shapes[shapeNumber].boundaryList, shapes[shapeNumber].freeList)).astype(int)
    nodeForce = []
    for node in nodes:
        nodeVectorField = vectorField(par, np.array([x[node], y[node]]), True, node)
        if solver.name == "electrostatic":
            if shapes[shapeNumber].material == 1:
                nodeForce.append(nodeVectorField*nodeSource[node])
            else:
                nodeForce.append(nodeVectorField*inducedNodeSource[node]*solver.physicalConstant)
        elif solver.name == "magnetostatic":
            thr = 90
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            if shapes[shapeNumber].material == 1:
                nodeForce.append(np.dot(R,nodeVectorField)*nodeSource[node])
            else:
                nodeForce.append(np.dot(R,nodeVectorField)*inducedNodeSource[node]/solver.physicalConstant)

    Force = sum(nodeForce)
    return Force


from .Solver import Solver
import numpy as np

class Electrostatic(Solver):
    def __init__(self):
        Solver.__init__(self)
        self.name = "electrostatic"
        self.family = "electric"
        self.physicalConstant = 8.8541878128*10**-12
        self.proportionalityConstant = 1/(2*np.pi*self.physicalConstant)

    def getApproximateMatrices(self, mesh, shapes):
        x = mesh.x
        y = mesh.y
        tri = mesh.conmat
        nn = len(x)
        ne = len(tri[:,0])
        pc = self.proportionalityConstant
        A = np.zeros((nn,nn))
        b = np.zeros((nn,1))
        nodeSource = mesh.nodeSource
        for i in range(ne):
            
            # 0-1 edge
            k = 0
            l = 1
            Ex = 0
            Ey = 0
            for shape in shapes:
                if not shape.isClosed:
                    nodes = shape.boundaryList
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
                else:
                    nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 1-2 edge
            k = 1
            l = 2
            Ex = 0
            Ey = 0
            for shape in shapes:
                if not shape.isClosed:
                    nodes = shape.boundaryList
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
                else:
                    nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 2-0 edge
            k = 2
            l = 0
            Ex = 0
            Ey = 0
            for shape in shapes:
                if not shape.isClosed:
                    nodes = shape.boundaryList
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
                else:
                    nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

        self.Aa = A
        self.ba = b

    def getLinNumericalMatrix(self, mesh):
        x = mesh.x
        y = mesh.y
        tri = mesh.conmat
        nn = len(x)
        A = np.zeros((nn,nn))

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = (P+Q)/(4*ea)
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        self.AnLin = A.copy()

    def getNumericalMatrix(self, mesh):
        x = mesh.x
        y = mesh.y
        tri = mesh.conmat
        nn = len(x)
        A = np.zeros((nn,nn))
        er = mesh.elementMaterial

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = er[i]*(P+Q)/(4*ea)
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        self.An = A.copy()

    def getApproximateBoundaryMatrices(self, mesh, shapes):
        x = mesh.x
        y = mesh.y
        pc = self.proportionalityConstant

        # Boundary nodes
        for i in range(len(shapes)):
            if len(shapes[i].parent) == 0:
                gs = i
                break

        lbn = shapes[gs].boundaryList
        lbn = lbn.astype(int)
        nbn = len(lbn)

        # Initializing matrices
        A = np.zeros([nbn,nbn])
        b = np.zeros([nbn,1])

        nodeSource = mesh.nodeSource

        for i in range(nbn):
            
            Ex = 0
            Ey = 0
            for shape in shapes:
                if not shape.isClosed:
                    nodes = shape.boundaryList
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[lbn[i-1]] + x[lbn[i]])/2 - px
                            Ry = (y[lbn[i-1]] + y[lbn[i]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
                else:
                    nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[lbn[i-1]] + x[lbn[i]])/2 - px
                            Ry = (y[lbn[i-1]] + y[lbn[i]])/2 - py
                            Ex = Ex + nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            # Forward Segment
            rx = x[lbn[i]] - x[lbn[i-1]]
            ry = y[lbn[i]] - y[lbn[i-1]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[i-1, i-1] = A[i-1, i-1] +1
            A[i-1, i] = -1
            b[i-1]           = b[i-1] + Er*np.sqrt(rx**2+ry**2)

            # Backward Segment
            rx = x[lbn[i-1]] - x[lbn[i]]
            ry = y[lbn[i-1]] - y[lbn[i]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[i, i] = A[i, i] +1
            A[i, i-1] = -1
            b[i]           = b[i] + Er*np.sqrt(rx**2+ry**2)

        self.Aab = A
        self.bab = b

    def getIterBoundaryMatrices(self, mesh, shapes, bt):
        x = mesh.x
        y = mesh.y
        ep0 = self.physicalConstant
        pc = self.proportionalityConstant

        # Boundary nodes
        for i in range(len(shapes)):
            if len(shapes[i].parent) == 0:
                gs = i
                break

        lbn = shapes[gs].boundaryList
        lbn = lbn.astype(int)
        nbn = len(lbn)

        # Initializing matrices
        A = np.zeros([nbn,nbn])
        b = np.zeros([nbn,1])

        nodeSource = bt.copy()
        nodeSource[lbn] = 0
        for i in range(nbn):
            
            Ex = 0
            Ey = 0
            for shape in shapes:
                if not shape.isClosed:
                    nodes = shape.boundaryList
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[lbn[i-1]] + x[lbn[i]])/2 - px
                            Ry = (y[lbn[i-1]] + y[lbn[i]])/2 - py
                            Ex = Ex + nodeSource[node]*ep0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*ep0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2
                else:
                    nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
                    for node in nodes:
                        if nodeSource[node] != 0:
                            px = x[node]
                            py = y[node]
                            Rx = (x[lbn[i-1]] + x[lbn[i]])/2 - px
                            Ry = (y[lbn[i-1]] + y[lbn[i]])/2 - py
                            Ex = Ex + nodeSource[node]*ep0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + nodeSource[node]*ep0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            # Forward Segment
            rx = x[lbn[i]] - x[lbn[i-1]]
            ry = y[lbn[i]] - y[lbn[i-1]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[i-1, i-1] = A[i-1, i-1] +1
            A[i-1, i] = -1
            b[i-1]           = b[i-1] + Er*np.sqrt(rx**2+ry**2)

            # Backward Segment
            rx = x[lbn[i-1]] - x[lbn[i]]
            ry = y[lbn[i-1]] - y[lbn[i]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[i, i] = A[i, i] +1
            A[i, i-1] = -1
            b[i]           = b[i] + Er*np.sqrt(rx**2+ry**2)

        self.Aib = A
        self.bib = b

    def getMiniLinNumericalMatrix(self, minix, miniy, miniTri):
        x = minix
        y = miniy
        tri = miniTri
        nn = len(x)
        A = np.zeros((nn,nn))

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = (P+Q)/(4*ea)
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        self.miniAn = A.copy()

    def findVectorFields(self, mesh, solver):
        x = mesh.x
        y = mesh.y
        u = solver.u.copy()

        Ex = np.zeros([len(x),1])
        Ey = np.zeros([len(x),1])
        neiNodes = mesh.neiNodes

        for i in range(len(x)):
            nNodes = neiNodes[i]
            for node in nNodes:
                rx = x[node]-x[i]
                ry = y[node]-y[i]
                Ex[i] = Ex[i] + (u[i] - u[node]) * rx / (np.sqrt(rx**2 + ry**2))**2
                Ey[i] = Ey[i] + (u[i] - u[node]) * ry / (np.sqrt(rx**2 + ry**2))**2

        solver.Ex = Ex
        solver.Ey = Ey


import numpy as np
from scipy.linalg import solve

class Solver():
    def __init__(self):
        pass

    def solveApproximateMatricesSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.ua = u.copy()
        self.u = u.copy()

    def solveApproximateMatricesScipy(self, A, b):
        self.ua = solve(A, b)
        self.u = self.ua.copy()

    def solveApproximateMatricesNumpy(self, A, b):
        self.ua = np.linalg.lstsq(A, b)[0]
        self.u = self.ua.copy()

    def solveNumericalMatricesSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.un = u.copy()
        self.u = u.copy()

    def solveNumericalMatricesScipy(self, A, b):
        self.un = solve(A, b)
        self.u = self.un.copy()

    def solveNumericalMatricesNumpy(self, A, b):
        self.un = np.linalg.lstsq(A, b)[0]
        self.u = self.un.copy()

    def solveApproximateBoundaryMatricesSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.uab = u.copy()
        self.u = u.copy()

    def solveApproximateBoundaryMatricesScipy(self, A, b):
        self.uab = solve(A, b)
        self.u = self.uab.copy()

    def solveApproximateBoundaryMatricesNumpy(self, A, b):
        self.uab = np.linalg.lstsq(A, b)[0]
        self.u = self.uab.copy()

    def solveNumericalMatricesApproxBoundSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.unab = u.copy()
        self.u = u.copy()

    def solveNumericalMatricesApproxBoundScipy(self, A, b):
        self.unab = solve(A, b)
        self.u = self.unab.copy()

    def solveNumericalMatricesApproxBoundNumpy(self, A, b):
        self.unab = np.linalg.lstsq(A, b)[0]
        self.u = self.unab.copy()

    def solveNumericalMatricesApproxIterBoundSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.unib = u.copy()
        self.u = u.copy()

    def solveNumericalMatricesApproxIterBoundScipy(self, A, b):
        self.unib = solve(A, b)
        self.u = self.unib.copy()

    def solveNumericalMatricesApproxIterBoundNumpy(self, A, b):
        self.unib = np.linalg.lstsq(A, b)[0]
        self.u = self.unib.copy()

    def solveIterBoundaryMatricesSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.uib = u.copy()
        self.u = u.copy()

    def solveIterBoundaryMatricesScipy(self, A, b):
        self.uib = solve(A, b)
        self.u = self.uib.copy()

    def solveIterBoundaryMatricesNumpy(self, A, b):
        self.uib = np.linalg.lstsq(A, b)[0]
        self.u = self.uib.copy()
    
    def solveminiScipy(self, A, b):
        self.miniu = solve(A, b)

    def solveminiNumpy(self, A, b):
        self.miniu = np.linalg.lstsq(A, b)[0]
        
    def solveminiSadiku(self, A, ui, b, target):
        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.miniu = u.copy()


# Import packages
import joblib
from . import PlotUtils
from . import Mesh
from .Solvers import Solve

def mesh(par, meshingMethod = "triangle", maxAngle = "", maxArea = "", crse = 1, dpi = 300):
    if par == []:
        geometry = joblib.load("Data/geometry.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        geometry = joblib.load(stri)

    Mesh.prepareGeometryForMeshing(geometry)
    Mesh.generateNodes(geometry, meshingMethod, maxAngle, maxArea, crse)
    mesh = Mesh.createMesh(geometry)

    if dpi < 300:
        dpi = 300

    if par == []:
        joblib.dump(geometry, "Data/geometry.sav")
        joblib.dump(mesh, "Data/mesh.sav")
        PlotUtils.plotMesh(geometry, mesh, "Figures/2_Mesh.png", dpi)
    else: 
        stri = f'Data/geometry_{par}.sav'
        joblib.dump(geometry, stri)
        stri = f'Data/mesh_{par}.sav'
        joblib.dump(mesh, stri)
        stri = f'Figures/2_Mesh_{par}.png'
        PlotUtils.plotMesh(geometry, mesh, stri, dpi)


def assigedSources(par, dpi):
    if par == []:
        shapes = joblib.load("Data/geometry.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        shapes = joblib.load(stri)
        stri = f'Data/mesh_{par}.sav'
        mesh = joblib.load(stri)
        stri = f'Data/solver_{par}.sav'
        solver = joblib.load(stri)

    Mesh.assignSources(shapes, mesh)

    if par == []:
        joblib.dump(mesh, "Data/mesh.sav")
        PlotUtils.plotSources(mesh, solver, "Figures/3_Sources.png", dpi)
    else:
        stri = f'Data/mesh_{par}.sav'
        joblib.dump(mesh, stri)
        stri = f'Figures/3_Sources_{par}.png'
        PlotUtils.plotSources(mesh, solver, stri, dpi)
    

def assigedMaterials(par, dpi):
    if par == []:
        shapes = joblib.load("Data/geometry.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        shapes = joblib.load(stri)
        stri = f'Data/mesh_{par}.sav'
        mesh = joblib.load(stri)
        stri = f'Data/solver_{par}.sav'
        solver = joblib.load(stri)

    Mesh.assignMaterials(shapes, mesh)
    
    if par == []:
        joblib.dump(mesh, "Data/mesh.sav")
        PlotUtils.plotMaterials(mesh, solver, "Figures/4_Materials.png", dpi)
    else:
        stri = f'Data/mesh_{par}.sav'
        joblib.dump(mesh, stri)
        stri = f'Figures/4_Materials_{par}.png'
        PlotUtils.plotMaterials(mesh, solver, stri, dpi)


def solution(par, solutionName = "numerical", solverMethod = "Scipy", withDifference = False, withField = False, dpi = 300):
    if par == []:
        shapes = joblib.load("Data/geometry.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
    else:
        stri = f'Data/geometry_{par}.sav'
        shapes = joblib.load(stri)
        stri = f'Data/mesh_{par}.sav'
        mesh = joblib.load(stri)
        stri = f'Data/solver_{par}.sav'
        solver = joblib.load(stri)

    Solve.getMatrices(shapes, mesh, solver, solutionName)
    Solve.solve(shapes, mesh, solver, solutionName, solverMethod)
    if dpi < 300:
        dpi = 300
        
    showSolution(par, mesh, solver, solutionName, withField, dpi)
    if withDifference:
        showDifference(par, mesh, solver, solutionName, dpi)

    if par == []:
        joblib.dump(mesh, "Data/mesh.sav")
        joblib.dump(solver, "Data/solver.sav")
    else:
        stri = f'Data/mesh_{par}.sav'
        joblib.dump(mesh, stri)
        stri = f'Data/solver_{par}.sav'
        joblib.dump(solver, stri)


def showSolution(par, mesh, solver, solutionName, withField, dpi):
    if (solutionName == "approximate"):
        solution = solver.ua
        PlotUtils.plotSolution(par, mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numerical"):
        solution = solver.un
        PlotUtils.plotSolution(par, mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numericalApproxBound"):
        solution = solver.unab
        PlotUtils.plotSolution(par, mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numericalIterBound"):
        solution = solver.unib
        PlotUtils.plotSolution(par, mesh, solver, solution, solutionName, withField, dpi)

def showDifference(par, mesh, solver, solutionName, dpi):
    pass
